
from pyutilib.misc import Options
from FDT import FDT

#
# This is a specialization of the FDT heuristic for the
# set-cover problem.
#
class SetCoverFDT(FDT):

    def __init__(self, options):
        FDT.__init__(self, options)

    def run(self, instance):
        #
        # Compute the LP relaxation of the model instance
        #
        root_lp_value = self.process_root_lp(instance)
        instance.pprint()
        #
        # Generate a reduced instance, where variables with value less thatn
        # a proscribed tolerance are fixed to zero.
        #
        reduced_instance, varmap = self.eliminate_zero_variables(instance, self.options.tol)
        #reduced_instance.display()
        omega = reduced_instance.solution()
        #print "OMEGA",omega
        #print "VARMAP",varmap
        # initial pool
        pool = [Container(values=omega, nfixed=0, integer=self.is_integer(omega, self.options.tol))]
        n = len(omega)
        # iterate until all locations are fixed
        for i in range(n):
            tmppool=[]
            did_work=False
            for node in pool:
                if node.integer:
                    node.nfixed += 1
                    tmppool.append(node)
                    continue

                branchinglp = self.generate_branching_lp(node.nfixed+1, n, node.values, varmap, reduced_instance)
                results = self.solver_mngr.solve(branchinglp, opt=self.opt, tee=self.options.tee, timelimit=self.options.timelimit, options=" ".join(self.options.lpsolver_options))
                results.write()
                if self.bad_results(results):
                    print("Terminating due to error in results")
                    return None
                soln = results.solution()
                print(soln.variable.keys())
                if soln.variable.l[0].value > self.options.tol:
                    values = []
                    for i in xsequence(n):
                        values.append( soln.variable.x[i,0].value / soln.variable.l[0].value )
                    did_work=True
                    tmppool += [ Container(nfixed=node.nfixed+1, values=values, integer=self.is_integer(values, self.options.tol)) ]
                if soln.variable.l[1].value > self.options.tol:
                    values = []
                    for i in xsequence(n):
                        values.append( soln.variable.x[i,1].value / soln.variable.l[1].value )
                    did_work=True
                    tmppool += [ Container(nfixed=node.nfixed+1, values=values, integer=self.is_integer(values, self.options.tol)) ]
            if did_work:
                if (self.options.always_prune or len(pool) > n):
                    pool = self.prune(tmppool, n, omega)
                else:
                    pool = tmppool
            print(pool)
        print(pool)
        for item in pool:
            #BUG: reduced_instance.load(item.values)
            for i in xsequence(len(item.values)):
                reduced_instance.variable(i).value = item.values[i]
            cost = reduced_instance.cost()
            #print "HERE", item.values, cost, "ratio=%s" % (root_lp_value / cost)
        # TODO: final stuff

    def bad_results(self, results):
        soln = results.solution()
        if soln.variable.l[0].value + soln.variable.l[1].value < 0.75:
            print("Lambda values do not sum to a value greater than 0.75!")
            return True
        return False

    def prune(self, pool, n, omega):
        print("Pool "+str(pool))
        #
        # Create a packing LP model
        #
        model = Model()
        #
        model.V = RangeSet(1,n)
        #
        def xs_rule(i, model):
            return omega[i-1]
        model.xs = Param(model.V, initialize=xs_rule)
        #
        model.Good = RangeSet(1,len(pool))
        #
        def Sols_rule(i, u, model):
            return pool[i-1].values[u-1]
        model.Sols = Param(model.Good, model.V, initialize=Sols_rule)
        #
        model.lam = Var(model.Good, bounds=(0.0,1.0))
        #
        def Prunedom_rule(model):
            return summation(model.lam)
        model.Prunedom = Objective(rule=Prunedom_rule, sense=maximize)
        #
        def ConvComb_rule(u, model):
            return sum([ model.lam[i]*model.Sols[i,u] for i in model.Good]) <= model.xs[u]
        model.ConvComb = Constraint(model.V, rule=ConvComb_rule)
        #
        # Construct and solve this model
        #
        instance = model.create()
        results = self.solver_mngr.solve(instance, opt=self.opt, tee=self.options.tee, timelimit=self.options.timelimit, options=" ".join(self.options.solver_options))
        #
        # Create new pool using the nonzero lambdas
        #
        newpool = []
        soln = results.solution()
        #print "Root LP", omega
        #print "Pool", pool
        for i in soln.variable.lam:
            #print "Packing Info:", i, soln.variable.lam[i].value, pool[i-1].values
            if soln.variable.lam[i].value > 0.0:
                newpool.append( pool[i-1] )
        return newpool

    def generate_branching_lp(self, fix_var, n, xb, varmap, root):
        #
        model = Model()
        #
        model.fix_var = Param(initialize=fix_var)
        #
        def SS_rule(model):
            ans = set()
            for (i,j) in root.S:
                try:
                    ans.add( (varmap[:(i-1)]+1, j) )
                except:
                    pass
            return ans
        model.SS = Set(dimen=2, rule=SS_rule)
        #
        model.L = Set(initialize=set([0,1]))
        #
        model.V = RangeSet(1,n)
        #
        model.Fixed = RangeSet(1,fix_var-1)
        #
        def xb_rule(i, model):
            return xb[i-1]
        model.xb = Param(model.V, rule=xb_rule)
        #
        def M_rule(model):
            ans = set()
            for (i,j) in model.SS:
                ans.add(j)
            return ans
        model.M = Set(rule=M_rule)
        #
        model.x = Var(model.V, model.L, bounds=(0.0,1.0))
        #
        model.l = Var(model.L, bounds=(0.0,1.0))
        #
        def Packdom_rule(model):
            return model.l[0]+model.l[1]
        model.Packdom = Objective(rule=Packdom_rule, sense=maximize)
        #
        def setcover_rule(i, mx, instance):
            expr = 0
            for (j,ii) in instance.SS:
                if ii == i:
                    expr += instance.x[j,mx]
            return instance.l[mx] <=  expr
        model.setcover = Constraint(model.M, model.L, rule=setcover_rule)
        #
        def upbound_rule(i, mx, instance):
            return instance.x[i,mx] <= instance.l[mx]
        model.upbound = Constraint(model.V, model.L, rule=upbound_rule)
        #
        def packinxs_rule(i, instance):
            return instance.x[i,0] + instance.x[i,1] <= instance.xb[i]
        model.packinxs = Constraint(model.V, rule=packinxs_rule)
        #
        def FixLoworHigh_rule(u, j, instance):
            return instance.x[u,j] == instance.l[j] * instance.xb[u]

        model.FixLoworHigh = Constraint(model.Fixed, model.L, rule=FixLoworHigh_rule)
        #
        def FixVarLoworHigh_rule(j, instance):
            return instance.x[value(instance.fix_var), j] == instance.l[j]*j
        model.FixVarLoworHigh = Constraint(model.L, rule=FixVarLoworHigh_rule)
        #
        # Create an instance of this model
        #
        instance = model.create()
        instance.pprint()
        return instance


if __name__ == "__main__":
    #
    # A hack to generate options
    #
    options=Options()
    #print 'YY'
    options._name = 'options'
    #print 'YY'
    options.keep_files=False
    options.log=False
    options.debug=True
    options.timelimit=None
    options.tee=True
    options.solver='cplex'
    options.solver_options=[]
    options.smanager_type='serial'

    #import sc
    #instance = sc.model.create()
    #fdt = SetCoverFDT(options)
    #fdt.run(instance)
